/*
 * cp40.i --
 *
 *	Interpreted part of the Yorick interface to CP40 data files and
 *	data processing.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-2006 Eric Thibaut.
 *
 *	This file is part of CP40soft.
 *
 *	CP40soft is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	CP40soft is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with CP40soft; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id$
 *	$Log$
 */

if (is_func(plug_in)) plug_in, "cp40";

extern cp40_new;
/* DOCUMENT id = cp40_new();
       -or- id = cp40_new(name);
       -or- id = cp40_new(name, mode);
     Creates a new CP40 data instance and get its identifier.  ID is needed
     for further reference to this CP40 data instance.  See cp40_assoc for
     the meaning of optional arguments NAME and MODE.  The data instance is
     deleted when the number of references to ID becomes zero.  When this
     occurs, the data stream associated with the CP40 instance, if any, is
     closed and updated if this is relevant (i.e. stream is seekable and its
     mode is "write" or "update").  If you don't like this behaviour, do:
       cp40_close, id, 0;
       id = [];

   SEE ALSO cp40_close, cp40_assoc, cp40_get, cp40_set. */

extern cp40_assoc;
/* DOCUMENT cp40_assoc(id, name, mode);
       -or- cp40_assoc(id, name);
     Associates CP40 data instance ID with file stream NAME.  Optional MODE
     is one of "r" or "read" (default), "w" or "write", "u" or "update".
     In read/update mode, you may want to check whether existing file is a
     valid CP40 data file with cp40_check(NAME).  The return value is ID.

   SEE ALSO cp40_check, cp40_new, cp40_close. */

func cp40_open(name, config=, window=, exposure=, spurious=, pixel=, parity=,
               correction=, scale=, angle=, raw_coefs=,
               flux=, spread=, distance=, ratio=)
/* DOCUMENT cp40_open(name)
       -or- cp40_open(name, key=value, ...)
     Opens  existing CP40 data  file NAME  in "read"  mode and  returns its
     object instance possibly after configuring some preprocessing options.

     Keyword  CONFIG can  be set  to the  name of  a configuration  file or
     another CP40  data file to  set preprocessing options of  the returned
     CP40 instance  (see cp40_copy).  Note: only  preprocessing options are
     affected by this keyword.

     Other preprocessing  attributes can be  set by using  keywords.  These
     settings always apply _after_ the CONFIG keyword.

     In  the following example,  the file  "HD222643_16.cp40" is  open with
     preprocessing   options  set   from  "eso-2000-07.cfg"   and  weighted
     distortion correction and a scale of 12 pixels per grid step:
       id = cp40_open("HD222643_16.cp40", config="ESO-2000-07.cfg",
                      correction="weighted", scale=12.0);

  KEYWORDS config,  window, exposure, spurious,  pixel, parity, correction,
           scale, angle, raw_coefs, flux, spread, distance, ratio.

  SEE ALSO cp40_new, cp40_copy, cp40_config. */
{
  id = cp40_new(name, "read");
  if (! is_void(config)) cp40_copy, config, id /*, preprocessing=1*/;
  return cp40_config(id, window=window, exposure=exposure, spurious=spurious,
                     pixel=pixel, parity=parity, correction=correction,
                     scale=scale, angle=angle, raw_coefs=raw_coefs,
                     flux=flux, spread=spread, distance=distance, ratio=ratio);
}

extern cp40_close;
/* DOCUMENT cp40_close, id;
       -or- cp40_close(id);
     Closes stream of CP40 data instance ID.  If keyword UPDATE is explicitly
     set to 0, the associated stream header is not updated.  The default is to
     update associated stream header if this is relevant (i.e. stream is
     seekable and its mode is write or update).  When called as a function,
     the return value is ID.

   SEE ALSO cp40_new, cp40_assoc, cp40_get, cp40_set. */

extern cp40_rewind;
/* DOCUMENT cp40_rewind(id);
       -or- cp40_rewind, id;
     Go to the beginning of data part of CP40 data instance ID.  Returns its
     argument if called as a function, making possible to do:
       count = cp40_run(cp40_rewind(id));

   SEE ALSO cp40_read. */

extern cp40_readable;
extern cp40_rewindable;
extern cp40_updatable;
extern cp40_writable;
/* DOCUMENT cp40_readable(id);
       -or- cp40_rewindable(id);
       -or- cp40_updatable(id);
       -or- cp40_writable(id);
     Check various capabilities of CP40 instance ID.

   SEE ALSO cp40_assoc, cp40_check. */

extern cp40_check;
/* DOCUMENT cp40_check(name);
     Returns non-zero if file NAME is a valid CP40 data file.

   SEE ALSO cp40_assoc, cp40_readable. */

extern cp40_get;
/* DOCUMENT cp40_get(id, name);
     Returns string value(s) for attribute NAME of CP40 data instance ID.
     VALUE will have the same geometry as NAME.  Non-existent attributes will
     result in nil values.

   SEE ALSO cp40_set, cp40_extra. */

extern cp40_set;
/* DOCUMENT cp40_set, id, name, value;
     Set string VALUE(s) for attribute NAME of CP40 data instance ID.
     VALUE must have the same geometry as NAME.  You can delete an
     extra-attribute (i.e. a non standard one) by setting its value to
     string(0).

   SEE ALSO cp40_get, cp40_extra. */

extern cp40_copy;
/* DOCUMENT cp40_copy, src, dst;
     Copy attributes of SRC to DST.  At least one of SRC or DST must be
     a CP40 data instance, the other can be either a CP40 data instance
     or a file name (scalar string) into which attributes are saved (DST
     is a string) or from which attributes are restored (SRC is a string).

     Set keywords ALL, EXTRA, PREPROCESSING and/or STANDARD to a non-nil
     non-zero value to copy/restore/save all, extra (non-standard),
     preprocessing and/or standard attributes respectively.  The default
     is to copy only preprocessing attributes.

   SEE ALSO cp40_attr, cp40_get, cp40_set. */

extern cp40_coefs;
/* DOCUMENT coef = cp40_coefs(id);
       -or- coef = cp40_coefs(id, coef);
     Respectively: get and set raw (unscaled) distortion correction
     coefficients, a (10,2,4) array of double, for CP40 data instance ID.

     If keyword SCALED is set to non-zero value, the effective (scaled)
     distortion correction coefficients are returned (there is no way to
     set directly these coefficients).

   SEE ALSO cp40_set, cp40_get, cp40_angle, cp40_scale. */

extern __cp40_reweight;
func cp40_reweight(this, w0, w1, w2, w3, reset=)
/* DOCUMENT cp40_reweight(this, w0, w1, w2, w34)
     Apply  flat-field  correction  by re-weighting  distortion  correction
     table(s) for  instance THIS.  W0, W1,  W2, and W3 are  the weights for
     channels 0, 1, 2, and 3 respectively.  Any of the weighting arrays can
     be nil (in  which case it is ignored).  The  dimensions of the non-nil
     weighting  arrays  must match  those  of the  CCD  for  the raw  pixel
     resolution as set in THIS.  In fact, the W's should have been computed
     from long exposure images computed with correction="none".

     If keyword  RESET is true,  all tasks in  THIS get cancelled,  THIS is
     rewinded and its correction method is set to "weighted".

     Returns THIS when called as a function.

   SEE ALSO cp40_coef, cp40_angle, cp40_scale. */
{
  if (reset) {
    cp40_cancel, this, "*";
    cp40_rewind, this;
    cp40_set, this, "correction", "weighted";
  }
  if (! is_void(w0)) __cp40_reweight, this, 0, w0;
  if (! is_void(w1)) __cp40_reweight, this, 1, w1;
  if (! is_void(w2)) __cp40_reweight, this, 2, w2;
  if (! is_void(w3)) __cp40_reweight, this, 3, w3;
  return this;
}

extern cp40_scale;
/* DOCUMENT scale= cp40_scale(id);
       -or- scale= cp40_scale(id, scale);
     Respectively: get and set scale (in pixels per calibration grid steps)
     of the distortion correction for CP40 data instance ID.

   SEE ALSO cp40_set, cp40_get, cp40_angle, cp40_coefs. */

extern cp40_angle;
/* DOCUMENT angle= cp40_angle(id);
       -or- angle= cp40_angle(id, angle);
     Respectively: get and set angle (in degrees anti-clockwise) of the
     distortion correction for CP40 data instance ID.  When setting this
     value, it may be useful to get the actual angle since it may be
     modified by the CP40 API, e.g. angles get wrapped into (-180,180].

   SEE ALSO cp40_set, cp40_get, cp40_scale, cp40_coefs. */

extern cp40_nframes;
extern cp40_nsaturations;
extern cp40_frame_count;
extern cp40_frame_date;
/* DOCUMENT nframes=      cp40_nframes(id);
       -or- nsaturations= cp40_nsaturations(id);
       -or- frame_count=  cp40_frame_count(id);
       -or- frame_date=   cp40_frame_date(id);
     Get current value of some read-only CP40 attributes.

   SEE ALSO cp40_set, cp40_get. */

extern cp40_exposure;
/* DOCUMENT exposure= cp40_exposure(id);
       -or- exposure= cp40_exposure(id, exposure);
     Respectively: get and set exposure (in number of short exposure frames)
     for super-frames.

   SEE ALSO cp40_set, cp40_get, cp40_scale, cp40_coefs. */

extern cp40_nphotons;
/* DOCUMENT nphotons= cp40_nphotons(id);
     Return an array of long's with the total number of photon for every
     channel of CP40 data instance ID.  BEWARE that CP40 channel indexing
     starts at 0 whereas Yorick indexing starts at 1... now you know!

   SEE ALSO cp40_set, cp40_get. */

/*---------------------------------------------------------------------------*/
/* REAL-TIME TASKS */

extern cp40_schedule;
/* DOCUMENT cp40_schedule(id, token, ...);
     Schedule photon-counting tasks identified by TOKEN (an array of strings)
     for CP40 instance ID.  There may be as many token arguments as needed.
     Nil strings are ignored.  ID is returned.

   SEE ALSO cp40_cancel, cp40_run, cp40_task_data. */

extern cp40_tasks;
/* DOCUMENT cp40_tasks(id);
     Return an array of strings with tokens of scheduled tasks for CP40
     instance ID.

   SEE ALSO cp40_schedule. */

extern cp40_cancel;
extern cp40_clear;
extern cp40_start;
extern cp40_stop;
/* DOCUMENT cp40_cancel(id, token, ...);
       -or- cp40_clear(id, token, ...);
       -or- cp40_start(id, token, ...);
       -or- cp40_stop(id, token, ...);
     Perform some operation onto photon-counting tasks identified by TOKEN
     (an array of strings) for CP40 instance ID.  There may be as may token
     arguments as needed.  Nil string are ignored and "*" stands for
     all-tasks.  ID is returned.

     cp40_cancel - cancel given task(s) and free all related resources.
     cp40_clear  - set to zero given task(s) data.
     cp40_start  - start a stopped photon-counting task(s).
     cp40_stop   - stop given running photon-counting task(s).

   SEE ALSO cp40_schedule, cp40_run, cp40_task_data. */

extern cp40_run;
/* DOCUMENT count = cp40_run(id);
       -or- count = cp40_run(id, maxcount);
     If MAXCOUNT is given, read and process as explained below at most that
     number of frames (not accounting for discarded frames); otherwise,
     read and process all remaining frames.  The number of processed frames
     (not accounting for discarded frames) is returned.

     For each frame, the processing consists in:
       1. reading next raw (super-)frame;
       2. apply preprocessing;
       3. update statistics for real-time;
       4. update real-time field-of-view image;
       5. stole raw (super-)frame (this is needed for circular averaging
          involved in real-time operation).

   SEE ALSO cp40_schedule, cp40_cancel_tasks, cp40_task_data. */

extern cp40_task_data;
/* DOCUMENT a = cp40_task_data(id, token);
     Returns task data identified by string TOKEN for CP40 instance ID.  The
     origins of the resulting array depends on the window settings for ID.

   SEE ALSO cp40_schedule, cp40_cancel_tasks, cp40_run,
            cp40_task_limits, orgsof, use_origins. */

extern cp40_task_limits;
/* DOCUMENT [xmin, xmax, ymin, ymax] = cp40_task_limits(id, token);
     Return 2-D limits for task data identified by string TOKEN for CP40
     instance ID.

   SEE ALSO cp40_task_data. */


/*---------------------------------------------------------------------------*/
extern cp40_stat;
/* DOCUMENT stat = cp40_stat(id);
       -or- stat = cp40_stat(id, count);

     Returns statistics related to frames of CP40 data ID for every CHANNEL,
     a 7-by-4-by-N array:

        STAT(1,C,)= frame date (an integer)
        STAT(2,C,)= flux
        STAT(3,C,)= photocentre abscissae
        STAT(4,C,)= photocentre ordinates
        STAT(5,C,)= spread (sigma) of photons
        STAT(6,C,)= distance of photocentre to nearest window border
        STAT(7,C,)= the ratio distance / spread.

     The optional parameter COUNT gives the number of frames to read (the
     actual number N of frames may be smaller because of end-of-data).
     If COUNT is zero, only the current frame is considered.  Otherwise,
     data starts with next data frame.

   SEE ALSO cp40_run. */

/*---------------------------------------------------------------------------*/
/* MANAGEMENT OF ATTRIBUTES */

func cp40_limits(id, chn1, chn2)
/* DOCUMENT limits = cp40_limits(id);
       -or- limits = cp40_limits(id, chn);
       -or- limits = cp40_limits(id, chn1, chn2);

     Get the limits, i.e. XMIN, XMAX, YMIN, YMAX (inclusive), for photon-
     event positions after pre-processing from CP40 data instance ID.

     Calls with one argument return the limits for all the channels:
       LIMITS(,1) gives [XMIN, XMAX, YMIN, YMAX] for channel 0 (not 1!)
       LIMITS(,2) gives [XMIN, XMAX, YMIN, YMAX] for channel 1 (not 2!)
       ...
     Calls with two arguments return the limits for channel CHN1 only.
     Calls with three arguments return the limits for the intercorrelation
     (autocorrelation if CHN1=CHN2) between channels CHN1 and CHN2.

   SEE ALSO: cp40_set, cp40_get. */
{
  l = array(long, 4, 4);
  if (sread(cp40_get(id, "extreme"), l) != 16) error, "BUG found";
  if (! is_void(chn2)) {
    l2 = l(,chn2+1);
    l1 = l(,chn1+1);
    return [l2(1)-l1(2), l2(2)-l1(1), l2(3)-l1(4), l2(4)-l1(3)];
  }
  if (! is_void(chn1)) return l(,chn1+1);
  return l;
}

func cp40_stds(nil)
/* DOCUMENT cp40_stds()
     Return names of standard CP40 attributes.

   SEE ALSO cp40_extra. */
{
  if (! is_void(nil)) error, "unexpected non-nil argument";
  return ["nframes", "nphotons", "nsaturations", "frame_count",
          "frame_date", "window", "extreme", "exposure", "spurious",
          "pixel", "correction", "scale", "angle", "raw_coefs",
          "coefs", "parity", "flux", "spread", "distance", "ratio"];
}

extern cp40_extra;
/* DOCUMENT names = cp40_extra(id);
     Get the names of all extra (non-standard) attributes of CP40 data
     instance ID.

   SEE ALSO cp40_set, cp40_get. */

func cp40_attr(id, name, value, extra=, channel=)
/* DOCUMENT cp40_attr -- Get/set CP40 attribute names and/or values.

     (1) Getting attribute names of CP40 data instance ID:
           name = cp40_attr(id);
	 where returned NAME is a string array of all attribute names (by
	 default).  To get only extra (i.e. non-standard) attribute names,
	 use keyword EXTRA with a non-zero value.  To get only standard
	 attribute names, use keyword EXTRA set to 0.

     (2) Getting attribute value for ID:
           value = cp40_attr(id, name);

     (3) Set attribute value and, possibly, get new value:
           cp40_attr, id, name, value;
	   newvalue = cp40_attr(id, name, value);

     Use keyword CHANNEL to appropriately append channel number (from 0
     to 3) to NAME.

   SEE ALSO cp40_set, cp40_get, cp40_config. */
{
  if (is_void(name)) {
    /* Return attribute names. */
    if (! is_void(value)) error, "VALUE with no NAME";
    if (is_void(extra)) {
      /* Return all attribute names. */
      result = cp40_stds();
      grow, result, cp40_extra(id);
      return result;
    }
    return (extra ? cp40_extra(id) : cp40_stds());
  }

  if (! is_void(channel)) name += swrite(format=",%d", channel);

  if (! is_void(value)) {
    if ((t = structof(value)) != string) {
      /* Convert value to string(s). */
      if (t == int || t == long || t == short || t == char) {
	value = swrite(format="%d", value);
      } else if (t == double) {
	value = swrite(format="%.16e", value);
      } else if (t == float) {
	value = swrite(format="%.7e", value);
      } else {
	error, "bad data type for VALUE";
      }
    }
    if ((n = numberof(name)) != numberof(value)) {
      if (n != 1) error, "non-conformable NAME and VALUE";
      value = cp40_strpack(value);
    }
    cp40_set, id, name, value;
  }
  if (! am_subroutine()) return cp40_get(id, name);
}

func cp40_config(id, window=, exposure=, spurious=, pixel=, parity=,
                 correction=, scale=, angle=, raw_coefs=,
                 flux=, spread=, distance=, ratio=)
/* DOCUMENT cp40_config(id, key=value, ...)
     Configure preprocessing attributes of CP40 instance ID using keywords
     syntax.  Implemented keywords are: WINDOW, EXPOSURE, SPURIOUS, PIXEL,
     PARITY, CORRECTION, SCALE, ANGLE, RAW_COEFS, FLUX, SPREAD, DISTANCE,
     and RATIO.  The return value is ID to allow for chained operations,
     e.g.:
       cp40_integ, cp40_rewind(cp40_config(id, exposure=1)), "ic01";

  SEE ALSO cp40_attr, cp40_set, cp40_strpack. */
{
  if (!is_void(exposure))   cp40_set,id,"exposure",  cp40_strpack(exposure);
  if (!is_void(spurious))   cp40_set,id,"spurious",  spurious;
  if (!is_void(pixel))      cp40_set,id,"pixel",     pixel;
  if (!is_void(parity))     cp40_set,id,"parity",    parity;
  if (!is_void(correction)) cp40_set,id,"correction",correction;
  if (!is_void(scale))      cp40_set,id,"scale",     cp40_strpack(scale);
  if (!is_void(angle))      cp40_set,id,"angle",     cp40_strpack(angle);
  if (!is_void(raw_coefs))  cp40_set,id,"raw_coefs", cp40_strpack(raw_coefs);
  if (!is_void(flux))       cp40_set,id,"flux",      cp40_strpack(flux);
  if (!is_void(spread))     cp40_set,id,"spread",    cp40_strpack(spread);
  if (!is_void(distance))   cp40_set,id,"distance",  cp40_strpack(distance);
  if (!is_void(ratio))      cp40_set,id,"ratio",     cp40_strpack(ratio);
  if (!is_void(window))     cp40_set,id,"window",    cp40_strpack(window);
  return id;
}

func cp40_info(id, names, values)
/* DOCUMENT s = cp40_info(id);
       -or- cp40_info, id;
     Returns a vector of strings describing all attribute names and values
     of CP40 data instance ID.  Just print out the result when called as
     a subroutine.

   SEE ALSO: cp40_extra, cp40_get. */
{
  stds= ["nframes", "nphotons,", "nsaturations",
	 "frame_count", "frame_date",
	 "window,", "extreme,",
	 "exposure", "spurious", "pixel",
	 "correction", "scale", "angle", "raw_coefs", "coefs",
	 "parity", "flux,", "spread,", "distance,", "ratio,"];
  one_per_channel = (strpart(stds, 0:0) == ",");
  names = array(string, numberof(stds) + 3 * numberof(where(one_per_channel)));
  j = 1;
  n = numberof(stds);
  for (i=1; i<=n; ++i) {
    if (one_per_channel(i)) {
      fmt= stds(i)+"%d";
      for (c=0; c<4; ++c) names(j++) = swrite(format=fmt, c);
    } else {
      names(j++)= stds(i);
    }
  }
  grow, names, cp40_extra(id);
  values = cp40_get(id, names);

  if (am_subroutine()) {
    write, format="%20s= %s\n", names, values;
  } else {
    return swrite(format="%20s= %s", names, values);
  }
}

func cp40_to_hash(id, db)
/* DOCUMENT cp40_to_hash(id)
       -or- cp40_to_hash, id, db;
     Copies most header information of CP40 instance ID in a hash table and
     returns it.  If  a second argument is provided, DB  is the target hash
     table; otherwise a new hash table is created.  As far as possible, the
     routine attempts to convert string values into numeric types.

   SEE ALSO h_new, cp40_info, ydf_save, ydf_restore. */

{
  if (is_void(db)) db = h_new();
  else if (! is_hash(db)) error, "expecting a hash table or nil";

  /* Get all the attributes of the CP40 instance but some unwanted/special
     ones. */
  opts = cp40_attr(id);
  n = numberof(opts);
  list = h_new();
  for (i=1 ; i<=n ; ++i) h_set, list, opts(i), [];
  h_delete, list, "scale", "angle", "raw_coefs", "coefs", "exposure",
    "nphotons", "nframes", "nsaturations",
    "frame_count", "frame_date";

  /* Set special members. */
  h_set, db,
    scale=cp40_scale(id),
    angle=cp40_angle(id),
    raw_coefs=cp40_coefs(id),
    coefs=cp40_coefs(id, scaled=1),
    nphotons=cp40_nphotons(id),
    nframes=cp40_nframes(id),
    nsaturations=cp40_nsaturations(id),
    exposure=cp40_exposure(id);
  l = array(long, 4, 4); // must be (re)created to avoid unwanted sharing
  if (sread(cp40_attr(id,"extreme"), l) == 16) {
    h_set, db, extreme=l;
    h_pop, list, "extreme";
  }
  l = array(long, 4, 4); // must be (re)created to avoid unwanted sharing
  if (sread(cp40_attr(id,"window"), l) == 16) {
    h_set, db, window=l;
    h_pop, list, "window";
  }

  /* Set other members (should do something for "dev*" members). */
  keys = h_keys(list);
  list = [];
  n = numberof(keys);
  for (i=1 ; i<=n ; ++i) h_set, db, keys(i), cp40_get(id, keys(i));

  return db;
}

/*---------------------------------------------------------------------------*/
/* INTEGRATION OF ESTIMATORS */

func cp40_integ(id, task, .., cancel=,
                window=, exposure=, spurious=, pixel=, parity=,
                correction=, scale=, angle=, raw_coefs=,
                flux=, spread=, distance=, ratio=)
/* DOCUMENT cp40_integ(id, task, ...)
       -or- cp40_integ, id, task, ...;
     Integrate photon  counting estimators given by TASK  (...), from _all_
     data  of instance  ID.  When  called as  a function,  the result  is a
     vector of pointers:
       *RESULT(i) = cp40_task_data(ID, TASK(i))
     where TASK(i) is the i-th scheduled task, with:
       TASK=cp40_tasks(ID).

   KEYWORDS
     CANCEL: cancel existing tasks; otherwise no task must have
             been scheduled for ID on entry.
     WINDOW, EXPOSURE, SPURIOUS, PIXEL, PARITY,
     CORRECTION, SCALE, ANGLE, RAW_COEFS, FLUX,
     SPREAD, DISTANCE, and RATIO: see cp40_config.

   SEE ALSO cp40_schedule, cp40_run, cp40_tasks,
            cp40_config, cp40_integ2. */
{
  if (is_array(cp40_tasks(id))) {
    if (cancel) cp40_cancel, id, "*";
    else error, "scheduled tasks already exist";
  }
  cp40_config, id, window=window, exposure=exposure, spurious=spurious,
    pixel=pixel, parity=parity, correction=correction, scale=scale,
    angle=angle, raw_coefs=raw_coefs, flux=flux, spread=spread,
    distance=distance, ratio=ratio;
  while (more_args()) grow, task, next_arg();
  cp40_schedule, id, (numberof(task) > 1 ? task(::-1) : task);
  cp40_run, cp40_rewind(id);
  if (! am_subroutine()) {
    tasks= cp40_tasks(id);
    n= numberof(tasks);
    result= array(pointer, n);
    for (i=1 ; i<=n ; ++i) {
      result(i)= &cp40_task_data(id, tasks(i));
    }
    return result;
  }
}

func cp40_integ2(id, .., cancel=, chunk=,
                 window=, exposure=, spurious=, pixel=, parity=,
                 correction=, scale=, angle=, raw_coefs=,
                 flux=, spread=, distance=, ratio=)
/* DOCUMENT cp40_integ2(id, task, ...)
       -or- cp40_integ2, id, task, ...;
     Integrate  photon counting  estimators and  corresponding  variance as
     given by TASK (...), from _all_  data of instance ID.  The result is a
     vector of pointers:
       *RESULT(1,i) = total estimator for TASKS(i)
       *RESULT(2,i) = variance for TASKS(i)
     where TASKS(i) is the i-th scheduled task, with TASKS = cp40_tasks(ID).

   KEYWORDS
     CANCEL: cancel existing tasks; otherwise no task must have
             been scheduled for ID on entry.
     CHUNK: number of frames per sample (default 50).
     WINDOW, EXPOSURE, SPURIOUS, PIXEL, PARITY,
     CORRECTION, SCALE, ANGLE, RAW_COEFS, FLUX,
     SPREAD, DISTANCE, and RATIO: see cp40_config.

   RESTRICTIONS
     One   should  always  look   at  cp40_tasks(ID)   to  get   the  exact
     correspondance between the result and the task.

     Due to  turbulence evolution, the estimated  variance strongly depends
     on the size of the subsets (value of CHUNK).

   SEE ALSO cp40_schedule, cp40_run, cp40_tasks,
            cp40_config, cp40_integ2. */
{
  local tasks;
  ntry_max = 100; // max # of failures to detect end-of-file
  if (is_void(chunk)) chunk = 50;

  if (is_array(cp40_tasks(id))) {
    if (cancel) cp40_cancel, id, "*";
    else error, "scheduled tasks already exist";
  }
  cp40_config, id, window=window, exposure=exposure, spurious=spurious,
    pixel=pixel, parity=parity, correction=correction, scale=scale,
    angle=angle, raw_coefs=raw_coefs, flux=flux, spread=spread,
    distance=distance, ratio=ratio;
  while (more_args()) grow, tasks, next_arg();
  cp40_schedule, id, (numberof(tasks) > 1 ? tasks(::-1) : tasks);
  tasks = cp40_tasks(id);
  ntasks = numberof(tasks);
  result = array(pointer, 2, ntasks);
  for (i=1 ; i<=ntasks ; ++i) {
    result(1,i) = &double(cp40_task_data(id, tasks(i)));
    result(2,i) = &double(cp40_task_data(id, tasks(i)));
  }
  cp40_rewind, id;

  nframes_total = 0;
  nframes = 0;
  nchunks = 0;
  ntry= 0;
  for (;;) {
    nframes+= cp40_run(id, chunk - nframes);
    if (nframes >= chunk) {
      if (nframes > chunk) error, "BUG in cp40_run";
      for (i=1 ; i<=ntasks ; ++i) {
        tmp = double(cp40_task_data(id, tasks(i)));
        if (min(tmp) < 0) error, "CHUNK too large (wrapping problem)";
        *result(1,i) += tmp;
        *result(2,i) += tmp*tmp;
      }
      cp40_clear, id, "*";
      nframes_total += nframes;
      nframes = 0;
      ++nchunks;
      ntry = 0;
      //if (nframes_total % 100 == 0) write, format=" %d", nframes_total;
    } else if (++ntry >= ntry_max) {
      scl = nchunks / (nchunks - 1.0);
      for (i=1 ; i<=ntasks ; ++i) {
        // FIXME: following expression does not preserve origins
        // tmp = scl * (*result(2,i) - (*result(1,i))*(*result(1,i))/nchunks);
        tmp = scl * (*result(2,i) - (*result(1,i))^2/nchunks);
        result(2,i) = &tmp;
      }
      write, format="\nNCHUNKS = %d\n", nchunks;
      write, format="\nNFRAMES = %d (%d skipped)\n", nframes_total, nframes;
      return result;
    }
  }
}

/*---------------------------------------------------------------------------*/

func cp40_ic2ps(ic, dx, dy, normalize=, dim=)
/* DOCUMENT cp40_ic2ps(ic)
       -or- cp40_ic2ps(ic, [dx, dy])
       -or- cp40_ic2ps(ic, dx, dy)
     Convert  intercorrelation  IC  into  powerspectrum.  (DX,DY)  are  the
     offsets (in  pixel units) of  the intercorrelation central  spike with
     respect to the first (lower-left)  pixel.  If no offsets are given, it
     is assumed that IC was computed with proper distortion correction such
     that the  center of  the intercorrelation is  at pixel (0,0)  when the
     origins  of array  IC  are  taken into  account  (see use_origins  and
     orgsof).

     If keyword  NORMALIZE is  true, the result  is normalized so  that its
     value at the zero-th frequency is 1.0.

     If keyword DIM is given, a DIM by DIM array is returned, otherwise the
     best dimension is guessed by fft_best_dim.

  SEE ALSO: fft_best_dim, fft_indgen.  */
{
  require, "fft_utils.i";
  { local var_ic; }

  /* Get dimensions. */
  dims = dimsof(ic);
  if (dims(1) != 2) error, "bad dimension list for IC";
  nx = dims(2);
  ny = dims(3);
  if (is_void(dim)) dim = fft_best_dim(max(nx,ny));

  /* Get offsets. */
  if (is_void(dx) && is_void(dy)) {
    /* Get the offset from the origins of the array. */
    orgs = cp40_orgsof(ic);
    dx = -orgs(2);
    dy = -orgs(3);
  } else {
    if (numberof(dx) == 2 && is_void(dy)) {
      dy = dx(2);
      dx = dx(1);
    } else if (numberof(dx) != 1 || dimsof(dx)(1) != 0 ||
               numberof(dy) != 1 || dimsof(dy)(1) != 0) {
      error, "expecting offsets as 2 scalars DX and DY or as [DX, DY]";
    }
    if ((s = structof(dx))!=double && s!=long && s!=int && s!=float) {
      error, "bad data type for DX";
    }
    if ((s = structof(dy))!=double && s!=long && s!=int && s!=float) {
      error, "bad data type for DY";
    }
  }

  /* Apply offsets. */
  if (dx < 0 || dx >= nx || dy < 0 || dy >= ny) error, "bad offsets";
  ps = array(complex, dim, dim);
  if (! (dx%1) && ! (dy%1)) {
    /* Integer offsets. */
    dx = long(dx);
    dy = long(dy);
    ps(1:nx-dx, 1:ny-dy) = ic(1+dx:0, 1+dy:0);
    if (dx) ps(1-dx:0,  1:ny-dy) = ic(1:dx, 1+dy:0);
    if (dy) ps(1:nx-dx, 1-dy:0) = ic(1+dx:0, 1:dy);
    if (dx && dy) ps(1-dx:0, 1-dy:0) = ic(1:dx, 1:dy);
    fft_inplace, ps, -1;
    ps = double(ps);
  } else {
    /* Apply phase tilt due to offset. */
    ps(1:nx, 1:ny) = ic;
    fft_inplace, ps, -1;
    PI = 3.14159265358979323846264338327950;
    u = (2.0*PI/dim) * fft_indgen(dim);
    v = u(-,);
    w = u*dx + v*dy;
    ps = cos(w)*double(ps) + sin(w)*(ps.im);
  }
  if (normalize) ps *= (1.0/ps(1,1));
  return ps;
}

/*---------------------------------------------------------------------------*/
/* OFFSETS */

extern cp40_shift;
/* DOCUMENT cp40_shift, id, channel, dx, dy;
     Modify raw  coefficients of instance  ID so that undistorted  image of
     CHANNEL  gets shifted  by offset  (DX,DY).  The  offsets are  given in
     undistorted pixel units  (i.e.  the routine takes care  of the current
     ANGLE and SCALE  attributes of INSTANCE).  ID is  returned when called
     as a function.

   WARNING After modifying the offsets, you should update the "window"
           attributes.

   SEE ALSO cp40_coefs, cp40_fit_offsets, cp40_fix_offsets. */

func cp40_fit_offsets(x, y, w)
/* DOCUMENT [dx, dy] = cp40_fit_offsets(x, y);
       -or- [dx, dy] = cp40_fit_offsets(x, y, w);
     Fit  CP40 channels  relative positions  given the  coordinates  of all
     intercorrelation spikes:
       [X(1), Y(1)] = coord. of spike in "ic01" = [X1-X0, Y1-Y0]
       [X(2), Y(2)] = coord. of spike in "ic02" = [X2-X0, Y2-Y0]
       [X(3), Y(3)] = coord. of spike in "ic03" = [X3-X0, Y3-Y0]
       [X(4), Y(4)] = coord. of spike in "ic12" = [X2-X1, Y2-Y1]
       [X(5), Y(5)] = coord. of spike in "ic13" = [X3-X1, Y3-Y1]
       [X(6), Y(6)] = coord. of spike in "ic23" = [X3-X2, Y3-Y2]

     Optional  third argument W  can be  used to  weight the  measures (all
     weights must be greater or equal zero).

     The result is the 3 by 2 array of offsets of channels 1, 2 and 3
     with respect to channel 0:
       [[X1-X0, X2-X0, X3-X0], [Y1-Y0, Y2-Y0, Y3-Y0]]

   SEE ALSO cp40_fix_offsets,  */
{
  if (numberof(x)!=6 || dimsof(x)(1)!=1) error, "bad abscissae X";
  if (numberof(y)!=6 || dimsof(y)(1)!=1) error, "bad ordinates Y";
  a = double([[1,0,0,-1,-1,0],[0,1,0,1,0,-1],[0,0,1,0,1,1]]);
  if (! is_void(w)) {
    if (numberof(w)!=6 || dimsof(w)(1)!=1) error, "bad weights W";
    w = sqrt(w);
    a *= w;
    x *= w;
    y *= w;
  }

  /* Solve the normal equations.  Note that when there are more equations
     than unknowns, QRsolve(a, b) is the same as
     LUsolve(a(+,)*a(+,), a(+,)*b(+)). */
  return QRsolve(a, [x, y]);
}

func cp40_fix_offsets(id, offsets)
/* DOCUMENT cp40_fix_offsets, id, offsets;
     Fix channel relative positions for instance ID given the array OFFSETS
     of intercorrelation spike positions as returned by cp40_fit_offsets.
   SEE ALSO cp40_fit_offsets, cp40_shift. */
{
  if ((s = structof(offsets)) != double && s != float &&
      s != long && s != int && s != short) error, "bad data type for OFFSETS";
  dims = dimsof(offsets);
  if (dims(1)!=2 || dims(2)!=3 || dims(3)!=2)
    error, "bad dimensions for OFFSETS";
  s = -offsets;
  cp40_shift, id, 1, s(1,1), s(1,2);
  cp40_shift, id, 2, s(2,1), s(2,2);
  cp40_shift, id, 3, s(3,1), s(3,2);
}

/*---------------------------------------------------------------------------*/

func cp40_all_its(id)
/* DOCUMENT cp40_all_its, id;
       -or- nframes= cp40_all_its(id);
     Compute _all_ integrations (long exposure images) for instance ID.
     Return number of selected frames.  To get the integration data, call
     cp40_task_data.  The following tasks are run onto the whole file (i.e.
     cp40_rewind is called before cp40_run):
       "it0", "it1", "it2", "it3".

   SEE ALSO cp40_all_ics, cp40_task_data, cp40_rewind,
            cp40_schedule, cp40_run. */
{
  if (numberof(cp40_tasks(id))) error, "there are scheduled tasks";
  cp40_schedule, cp40_rewind(id), swrite(format="it%d", indgen(0:3));
  return cp40_run(id);
}

func cp40_all_ics(id)
/* DOCUMENT cp40_all_ics, id;
       -or- nframes= cp40_all_ics(id);
     Compute _all_ intercorrelations for instance ID.  Return number of
     selected frames.  To get the intercorrelation data, call cp40_task_data.
     The following tasks are run onto the whole file (i.e. cp40_rewind is
     called before cp40_run):
       "ic01", "ic02", "ic03", "ic12", "ic13", "ic23".

   SEE ALSO cp40_all_its, cp40_task_data, cp40_rewind,
            cp40_schedule, cp40_run. */
{
  if (numberof(cp40_tasks(id))) error, "there are scheduled tasks";
  cp40_rewind, id;
  for (chn0=0 ; chn0<4 ; ++chn0) {
    for (chn1=chn0+1 ; chn1<4 ; ++chn1) {
      cp40_schedule, id, swrite(format="ic%d%d", chn0, chn1);
    }
  }
  return cp40_run(id);
}

func cp40_center_and_fit_offset(id, wmin=, wmax=, win=, unwrap=, size=)
{
  tasks= ["ic01", "ic02", "ic03", "ic12", "ic13", "ic23"];
  x = y = array(double, 6);
  for (i=1 ; i<=6 ; ++i) {
    xy = cp40_centroid_of_ic(id, tasks(i), orgs=1, unwrap=unwrap, size=size,
                             win=win, wmin=wmin, wmax=wmax);
    x(i) = xy(1);
    y(i) = xy(2);
  }
  return cp40_fit_offsets(x, y);
}

func cp40_centroid_of_ic(id, task, win=, wmin=, wmax=, orgs=, unwrap=, size=)
/* DOCUMENT cp40_centroid_of_ic(id, task)
       -or- cp40_centroid_of_ic(ic)
     Interactively   get   coordinates   [DX,DY]   of  speckle   spike   in
     intercorrelation with  respect to  the first (lower-left)  pixel.  The
     intercorrelation can be  either specified by a CP40  instance ID and a
     identifier TASK or as a 2D array (this only change the way coordinates
     are  displayed).  Keyword  WIN can  be  used to  specify a  particular
     window to display the image.

     Keyword WMIN can be used to specify the lowest wavelet order (starting
     at 1) to  keep; this is useful to remove  some high-frequency noise in
     the intercorrelation.  Keyword WMAX can be used to specify the highest
     wavelet order to keep; this  is useful to remove the smooth background
     (mainly due to  turbulence) in the intercorrelation.  If  none of WMIN
     and WMAX are set, no wavelet filtering is performed.

     Unless keyword ORGS is true, the returned value of cp40_centroid_of_ic
     is suitable for cp40_ic2ps; for instance:
        offset = cp40_centroid_of_ic(ic);
        ps = cp40_ic2ps(ic, offset(1), offset(2));

     If keyword ORGS is true, the origins of the intercorrelation array are
     taken into  account and the  returned value of  cp40_centroid_of_ic is
     suitable for cp40_fit_offsets.

   SEE ALSO cp40_task_data, yeti_wavelet, cp40_ic2ps, cp40_fit_offsets,
            cp40_center_and_fit_offset. */
{
  local ic;
  require, "yeti_fit_2d_spike.i";

  /* Number of wavelet orders to compute. */
  if (is_void(wmin)) wmin = 1;
  if (is_void(wmax)) {
    wnum = wmin;
  } else {
    wnum = wmax + 1;
  }

  if (is_array(id)) {
    orgs = use_origins(1); /* do use origins */
    orgs = orgsof(id);     /* overwrite ORGS -> restore previous origin mode */
    if (orgs(1) != 2) error, "expecting a 2D array";
    x0 = orgs(2);
    y0 = orgs(3);
    eq_nocopy, ic, id;
  } else {
    ic = cp40_task_data(id, task);
    lm = cp40_task_limits(id, task);
    x0 = lm(1);
    y0 = lm(3);
  }
  dims = dimsof(ic);
  if (dims(1) != 2) error, "bad dimension list for intercorrelation";
  if (unwrap) ic = cp40_unwrap(ic);
  if (wnum != 1) {
    /* Wavelet filtering. */
    ic = yeti_wavelet(ic, wnum)(.., sum:wmin:wmax);
  }
  nx = dims(2);
  ny = dims(3);
  x1 = nx - 1 + x0;
  y1 = ny - 1 + y0;
  if (is_void(win)) {
    win = current_window();
    if (win < 0) win = 7;
  }
  winkill, win;
  window, win, width=552, height=900, dpi=90;
  do_not_use_origins = use_origins(0);
  if (is_void(size)) size = 7;
  prm = gg_fit_2d_spike(ic, x0, y0, x1, y1, background=1, size=size);
  if (is_void(prm)) return [];
  if (orgs) return [prm(3), prm(4)];
  return [prm(3) - x0, prm(4) - y0];


#if 0 /* old code */
  if (is_void(radius)) radius= 5;
  if (is_void(wmin)) wmin= 2;
  if (is_void(wmax)) wmax= 3;
  if (is_void(niters)) niters= 2;
  l= cp40_task_limits(id, task);
  x= indgen(l(1):l(2));
  y= indgen(l(3):l(4))(-,);
  z= yeti_wavelet(cp40_task_data(id, task), wmax+1)(.., sum:wmin:wmax);
  r= abs(x, y);
  if (! is_void(visu)) {
    window, visu;
    fma;
    pli, z, l(1)-0.5, l(3)-0.5, l(2)+0.5, l(4)+0.5;
    limits;
    rdline, prompt="Zoom/pan, then hit [return]";
  }
  i= z(*)(mxx);
  x0= x(1+(i-1)%numberof(x));
  y0= y(1+(i-1)/numberof(x));
  while (--niters > 0) {
    write, format="centroid of %s: %.2f %.2f\n", task, double(x0), double(y0);
    r= abs(x-x0, y-y0);
    w= z * ((radius % 1) ? r <= radius : r < radius+0.5);
    sw= sum(w);
    x0= sum(w*x)/sw;
    y0= sum(w*y)/sw;
  }
  if (! am_subroutine()) return [x0, y0];
  write, format="centroid of %s: %.2f %.2f\n", task, x0, y0;
#endif
}

/*---------------------------------------------------------------------------*/
/* IMAGES */

func cp40_orgsof(a)
/* DOCUMENT cp40_orgsof(a)
     Get actual origins of array A (as if "use_origins" is active).

   SEE ALSO orgsof, use_origin, cp40_pli. */
{
  local_object = use_origins(1);
  return orgsof(a);
}

func cp40_pli(id, token, legend=, hide=, top=, cmin=, cmax=)
/* DOCUMENT cp40_pli, id, token;
       -or- cp40_pli, img;
     Plot data of task identified by string TOKEN for CP40 instance ID
     or array IMG taking care of setting the correct coordinates system.

   KEYWORDS: legend, hide, top, cmin, cmax.

   SEE ALSO pli, orgsof, use_origin, cp40_task_data, cp40_task_limits. */
{
  if (is_array(id) && is_void(token)) {
    orgs = cp40_orgsof(id);
    if (orgs(1) != 2) error, "expecting 2-D array";
    dims = dimsof(id);
    l = [orgs(2), orgs(2)+dims(2)-1, orgs(3), orgs(3)+dims(3)-1];
  } else if (strpart(token, 0:0)=="*") {
    token = strpart(token, :-1);
    if (token == "it") {
      for (i=0 ; i<4 ; ++i) {
	window, i;
	fma;
        logxy, 0, 0;
	cp40_pli, id, swrite(format="%s%d", token, i), legend=legend,
	  hide=hide, top=top, cmin=cmin, cmax=cmax;
	limits;
      }
      return;
    } else {
      error, "display of all \""+token+"\" not yet implemented";
    }
  } else {
    l = cp40_task_limits(id, token);
    id= cp40_task_data(id, token);
  }
  pli, id, l(1)-0.5, l(3)-0.5, l(2)+0.5, l(4)+0.5,
    legend=legend, hide=hide, top=top, cmin=cmin, cmax=cmax;
}

func cp40_unwrap(img)
/* DOCUMENT cp40_unwrap(img)
     Unwrap the image IMG.  This routines is an attempt to solve the problem
     due to the limited dynamic of data used to integrate estimators in
     CP40 librairy (a 32-bit integer).

   SEE ALSO cp40_schedule, cp40_run, cp40_integ, cp40_integ2. */
{
  step = 2.0^(8*sizeof(long));
  threshold= 0.3*step;
  if (1) {
    a = double(img); // also copy origins
    dim2 = dimsof(img)(3);
  } else {
    // FIXME: following statements does not preserve origins
    dim1= dimsof(img)(2);
    dim2= dimsof(img)(3);
    a= array(double, dim1, dim2);
  }
  for (i=1 ; i<=dim2 ; ++i) {
    line = double(img(,i));
    if (min(line) < 0) {
      grad = line(dif);
      k = where(grad <= -threshold);
      if (is_array(k)) {
        n = numberof(k);
        for (j=1 ; j<=n ; ++j) line(k(j)+1:) += step;
      }
      k = where(grad >= threshold);
      if (is_array(k)) {
        n = numberof(k);
        for (j=1 ; j<=n ; ++j) line(k(j)+1:) -= step;
      }
    }
    a(,i) = line;
  }
  return a;
  //return img + 2.0^32*(img<0);
}

/*---------------------------------------------------------------------------*/
/* STRING UTILITIES */

func cp40_strpack(x)
/* DOCUMENT cp40_strpack(x);
     Get a scalar string obtained by packing elements of X, possibly
     converted to string, separated by spaces.

   SEE ALSO cp40_strsplit, swrite. */
{
  if ((n = numberof(x)) == 0) return string(0);
  if ((t = structof(x)) != string) {
    if (t == int || t == long || t == short || t == char) {
      x = swrite(format="%d", x);
    } else if (t == double) {
      x = swrite(format="%.16e", x);
    } else if (t == float) {
      x = swrite(format="%.7e", x);
    } else {
      error, "bad data type";
    }
  }
  if (n == 1) return x;
  s = x(1);
  for (i=2 ; i<=n ; ++i) s += " " + x(i);
  return s;
}

func cp40_strsplit(s)
/* DOCUMENT cp40_strsplit(s);
     Get an array of strings constituted by the words of S (words are
     separated by one or several spaces or null characters).

   SEE ALSO cp40_strpack. */
{
  b = char(*pointer(s));
  w = where((b < '!') * (b >= '\0')); /* Must check that b >= '\0' since
                                         char may be signed... */
  if ((n = numberof(w)) > 0) {
    s = [];
    i2 = 0;
    for (i=1 ; i<=n ; ++i)
      if ((i1 = i2+1) < (i2 = w(i))) grow, s, string(&(b(i1:i2-1)));
  }
  return s;
}

/*---------------------------------------------------------------------------*/
